﻿using System.Collections;
using System.Reflection;

namespace Connected.Interop.Merging
{
	internal sealed class ObjectMerger : Merger
	{
		public void Merge(object destination, params object[] sources)
		{
			if (destination is null || !HasSource(sources))
				return;

			var sourceProperties = PropertyAggregator.Aggregate(sources);

			foreach (var property in Properties.GetImplementedProperties(destination))
				MergeProperty(destination, sourceProperties, property);
		}

		private static bool HasSource(params object[] sources)
		{
			foreach (var source in sources)
			{
				if (source is not null)
					return true;
			}

			return false;
		}

		private void MergeProperty(object destination, Dictionary<string, object> sourceProperties, PropertyInfo property)
		{
			if (property.PropertyType.IsTypePrimitive())
			{
				if (!property.CanWrite)
					return;

				if (!sourceProperties.TryGetValue(property.Name, out object? source))
					return;

				if (source.GetType() is Type propertyType && PropertyResolver.Resolve(propertyType, property.Name) is PropertyInfo propertyInfo)
					property.SetValue(destination, propertyInfo.GetValue(source));
			}
			else if (IsArray(property))
				MergeEnumerable(destination, sourceProperties, property);
			else
				MergeObject(destination, sourceProperties, property);
		}

		private void MergeEnumerable(object destination, Dictionary<string, object> sourceProperties, PropertyInfo property)
		{
			if (sourceProperties.TryGetValue(property.Name, out object? sourceProperty))
				return;

			if (property.GetValue(sourceProperty) is not IEnumerable sourceValue)
				return;

			var sourceElements = GetElements(sourceValue);
			var destinationValue = property.GetValue(destination);

			if (destinationValue is null && !property.CanWrite)
				return;

			var addMethod = property.PropertyType.GetMethod(nameof(IList.Add), BindingFlags.Public | BindingFlags.Instance | BindingFlags.NonPublic);
			var instance = addMethod is not null ? Activator.CreateInstance(property.PropertyType) : Array.CreateInstance(property.PropertyType, sourceElements.Count);
			var elementType = instance.GetType().GetElementType();

			for (var i = 0; i < sourceElements.Count; i++)
			{
				/*
			    * TODO: handle Dictionary
			    */
				var sourceElement = sourceElements[i];
				var item = Activator.CreateInstance(elementType);

				Merge(sourceElement, item);

				if (addMethod is not null)
					addMethod.Invoke(instance, new object[] { item });
				else
					((Array)instance).SetValue(item, i);
			}

			property.SetValue(destination, instance);
		}

		private void MergeObject(object destination, object source, PropertyInfo property)
		{
			throw new NotImplementedException();
		}
	}
}
